package org.bdware.sc.event;

import com.google.gson.JsonObject;
import com.google.gson.reflect.TypeToken;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.db.KeyValueRocksDBUtil;
import org.bdware.sc.event.REvent.REventSemantics;
import org.bdware.sc.event.REvent.REventType;
import org.bdware.sc.event.clients.ContractConsumer;
import org.bdware.sc.event.clients.IEventConsumer;
import org.bdware.sc.event.clients.IEventConsumer.ConsumerType;
import org.bdware.sc.event.clients.NodeConsumer;
import org.bdware.sc.util.HashUtil;
import org.bdware.sc.util.JsonUtil;

import java.lang.reflect.Type;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.concurrent.ConcurrentHashMap;

import static org.bdware.sc.event.REvent.REventType.SUBSCRIBE;
import static org.bdware.sc.event.REvent.REventType.UNSUBSCRIBE;

/**
 * @author Kaidong Wu
 */
public class EventRecorder {
    private static final Logger LOGGER = LogManager.getLogger(EventRecorder.class);
    private static final String LATEST_EVENT_KEY = "LATEST_EVENT", TEMP_TOPIC_KEYS = "TEMP_TOPICS";

    private final String dbName;
    private final EventBroker broker;
    private final RecordPointer latestEvent = new RecordPointer();

    public EventRecorder(String dbName, EventBroker broker) {
        this.dbName = dbName;
        this.broker = broker;
    }

    public void appendEvent(REvent e) {
        ETransaction transaction = new ETransaction(e);
        synchronized (latestEvent) {
            transaction.prev = latestEvent.get();
            String tranJson = "et" + JsonUtil.toJson(transaction);
            String key = HashUtil.sha3(tranJson);
            latestEvent.set(key);
            latestEvent.setCp(false);
            KeyValueRocksDBUtil.instance.setValue(dbName, LATEST_EVENT_KEY, latestEvent.get());
            KeyValueRocksDBUtil.instance.setValue(dbName, key, tranJson);
        }
    }

    public void createCheckPoint(Map<String, Set<String>> topic2cIds,
                                 Map<String, IEventConsumer> id2Consumers) {
        CheckPoint cp = new CheckPoint(topic2cIds, id2Consumers);
        synchronized (latestEvent) {
            if (!latestEvent.isCp()) {
                cp.prev = latestEvent.get();
                String cpJson = "cp" + JsonUtil.toJson(cp);
                String key = HashUtil.sha3(cpJson);
                latestEvent.set(key);
                latestEvent.setCp(true);
                KeyValueRocksDBUtil.instance.setValue(dbName, LATEST_EVENT_KEY, latestEvent.get());
                KeyValueRocksDBUtil.instance.setValue(dbName, key, cpJson);
            }
        }
    }

    /**
     * recover registry from database
     *
     * @return a check point, including topic registry and client registry
     */
    public CheckPoint recoverRegistryFromDb() {
        Stack<Object> stack = new Stack<>();
        // recover the latest event pointer from database
        String key = KeyValueRocksDBUtil.instance.getValue(dbName, LATEST_EVENT_KEY);
        latestEvent.set(key);
        CheckPoint cp = new CheckPoint();
        Type topic2cIdsType = TypeToken.getParameterized(ConcurrentHashMap.class, String.class,
                TypeToken.getParameterized(Set.class, String.class,
                        String.class).getType()).getType();
        // retrieving transactions from database
        while (null != key && !key.isEmpty()) {
            String json = KeyValueRocksDBUtil.instance.getValue(dbName, key);
            if (null == json || json.isEmpty()) {
                LOGGER.warn("record damaged! " + key);
                break;
            }
            try {
                if (json.startsWith("cp")) {
                    // create check point by the transaction and stop retrieving
                    JsonObject data = JsonUtil.parseStringAsJsonObject(json.substring(2));
                    cp.topic2cIds = JsonUtil.fromJson(data.get("topic2cIds").toString(), topic2cIdsType);
                    JsonObject id2Consumers = data.getAsJsonObject("id2Consumers");
                    for (String k : id2Consumers.keySet()) {
                        JsonObject consumer = id2Consumers.getAsJsonObject(k);
                        if (!consumer.has("type")) {
                            continue;
                        }
                        ConsumerType type = ConsumerType.valueOf(consumer.get("type").getAsString());
                        switch (type) {
                            case Contract:
                                cp.id2Consumers.put(k,
                                        new ContractConsumer(consumer.get("contract").getAsString(),
                                                consumer.get("handler").getAsString()));
                                break;
                            case Node:
                                cp.id2Consumers.put(k, new NodeConsumer(k));
                            default:
                                break;
                        }
                    }
                    for (String topic : cp.topic2cIds.keySet()) {
                        Set<String> idSet = cp.topic2cIds.get(topic);
                        for (String id : idSet) {
                            if (!cp.id2Consumers.containsKey(id)) {
                                cp.id2Consumers.remove(id);
                            }
                        }
                    }
                    cp.prev = data.get("prev").getAsString();
                    break;
                } else if (json.startsWith("et")) {
                    // push transactions into the stack
                    ETransaction tran = JsonUtil.fromJson(json.substring(2), ETransaction.class);
                    key = tran.prev;
                    // just sub or unsub event need to be processed
                    if (tran.type == SUBSCRIBE || tran.type == UNSUBSCRIBE) {
                        stack.push(tran);
                    }
                } else {
                    LOGGER.warn("record damaged! " + key);
                    break;
                }
            } catch (Exception ignored) {
                LOGGER.warn("record damaged! " + key);
                LOGGER.debug("record damaged! " + key + ": " + json);
                break;
            }
        }
        if (stack.isEmpty()) {
            // if empty, return the check point
            latestEvent.setCp(true);
        } else {
            // on the base of old check point, process following sub or unsub events to recover registry
            while (!stack.empty()) {
                Object record = stack.pop();
                if (record instanceof CheckPoint) {
                    cp = (CheckPoint) record;
                } else {
                    ETransaction tran = (ETransaction) record;
                    IEventConsumer consumer = broker.parseConsumer(tran.content);
                    switch (tran.type) {
                        case SUBSCRIBE:
                            if (!broker.subInReg(tran.topic, consumer, cp.topic2cIds, cp.id2Consumers)) {
                                LOGGER.warn("record damaged! " + key);
                                LOGGER.debug("record damaged! " + key + ": " + JsonUtil.toJson(tran));
                            }
                            break;
                        case UNSUBSCRIBE:
                            if (!broker.unsubInReg(tran.topic, consumer, cp.topic2cIds, cp.id2Consumers)) {
                                LOGGER.warn("record damaged! " + key);
                                LOGGER.debug("record damaged! " + key + ": " + JsonUtil.toJson(tran));
                            }
                        default:
                            break;
                    }
                }
            }
        }
        LOGGER.info("recover registry from database done!");
        return cp;
    }

    public void saveTempTopics(Map<String, Long> tempTopics) {
        KeyValueRocksDBUtil.instance.setValue(dbName, TEMP_TOPIC_KEYS, JsonUtil.toJson(tempTopics));
    }

    public Map<String, Long> recoverTempTopicsFromDb() {
        String json = KeyValueRocksDBUtil.instance.getValue(dbName, TEMP_TOPIC_KEYS);
        Map<String, Long> ret = new ConcurrentHashMap<>();
        try {
            JsonObject jo = JsonUtil.parseStringAsJsonObject(json);
            for (String key : jo.keySet()) {
                ret.put(key, jo.get(key).getAsLong());
            }
        } catch (Exception ignored) {
        }
        return ret;
    }

    static class ETransaction {
        REventType type;
        REventSemantics semantics;
        String topic;
        String content;
        String requestID;
        // public key of sender
        String sender;
        String signature;
        String prev;

        public ETransaction(REvent e) {
            type = e.getType();
            semantics = e.getSemantics();
            topic = e.getTopic();
            content = e.getContent();
            requestID = e.getRequestID();
            sender = e.getPublicKey();
            signature = e.getSignature();
        }
    }

    static class CheckPoint {
        Map<String, Set<String>> topic2cIds;
        Map<String, IEventConsumer> id2Consumers;
        String prev;

        public CheckPoint() {
            topic2cIds = new ConcurrentHashMap<>();
            id2Consumers = new ConcurrentHashMap<>();
        }

        public CheckPoint(Map<String, Set<String>> topic2cIds,
                          Map<String, IEventConsumer> id2Consumers) {
            this.topic2cIds = topic2cIds;
            this.id2Consumers = id2Consumers;
        }
    }

    static class RecordPointer {
        private String str;
        private boolean isCp = false;

        public String get() {
            return str;
        }

        public void set(String str) {
            this.str = str;
        }

        public boolean isCp() {
            return isCp;
        }

        public void setCp(boolean cp) {
            isCp = cp;
        }
    }
}
