package org.bdware.sc.event.clients;

import com.google.gson.annotations.Expose;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.ContractClient;
import org.bdware.sc.ContractManager;
import org.bdware.sc.ContractResult;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.util.HashUtil;
import org.bdware.sc.util.JsonUtil;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * @author Kaidong Wu
 */
public class ContractConsumer implements IEventConsumer {
    private static final Logger LOGGER = LogManager.getLogger(ContractConsumer.class);
    private static final int TIMEOUT_COUNT = 5;
    private static final Map<String, ScheduledFuture<?>> scheduledFutures =
            new ConcurrentHashMap<>();

    private final String contract;
    private final String handler;
    @Expose(serialize = false, deserialize = false)
    private final Object flag = new Object();
    @SuppressWarnings("unused")
    private final ConsumerType type = ConsumerType.Contract;

    public ContractConsumer(String contract, String handler) {
        this.contract = contract;
        this.handler = handler;
    }

    public String getContract() {
        return contract;
    }

    @Override
    public String getId() {
        return HashUtil.sha3(contract, handler);
    }

    @Override
    public ConsumerType getType() {
        return type;
    }

    @Override
    public void publishEvent(String msg, ResultCallback rc, String... options) {
        executeContract(msg, this.handler, rc, options);
    }

    @Override
    public void competeSub(String msg, ResultCallback rc, String... options) {
        executeContract(msg, "_preSub", rc, options);
    }

    private void executeContract(String msg, String handler, ResultCallback rc, String... options) {
        ContractRequest cr = new ContractRequest();
        if (options.length > 1 && null != options[1]) {
            cr.setRequester(options[1]);
        } else {
            cr.setRequester("event");
        }
        cr.setContractID(contract);
        cr.setAction(handler);
        cr.setArg(msg);
        cr.setRequestID(options[0]);
        ContractClient cc = ContractManager.instance.getClient(contract);
        if (null == cc) {
            rc.onResult((String) null);
            return;
        }
        AtomicInteger callCount = new AtomicInteger(0);
        ScheduledFuture<?> future = ContractManager.scheduledThreadPool.scheduleAtFixedRate(
                () -> ContractManager.instance.executeContractInternal(cr, new ResultCallback() {
                    @Override
                    public void onResult(String str) {
                        boolean ret = true;
                        try {
                            ContractResult result = JsonUtil.fromJson(str, ContractResult.class);
                            if (!result.status.equals(ContractResult.Status.Success)) {
                                if (callCount.get() == TIMEOUT_COUNT ||
                                        (result.status == ContractResult.Status.Exception &&
                                                result.result.toString().contains("not exported"))) {
                                    rc.onResult((String) null);
                                } else {
                                    ret = false;
                                }
                            }
                        } catch (Exception e) {
                            LOGGER.warn("receiving event error! " + contract + "." + handler + ": " + e.getMessage());
                            rc.onResult((String) null);
                        }
                        callCount.incrementAndGet();
                        if (ret) {
                            synchronized (flag) {
                                try {
                                    // wait for setting scheduledFutures
                                    flag.wait(500L);
                                } catch (InterruptedException e) {
                                    LOGGER.error(e.getMessage());
                                }
                            }
                            scheduledFutures.get(getId()).cancel(true);
                            scheduledFutures.remove(getId());
                        }
                    }
                }, (reqID, hashStr) -> {
                }),
                500L,
                2500L,
                TimeUnit.MILLISECONDS);
        scheduledFutures.put(getId(), future);
        synchronized (flag) {
            flag.notify();
        }
    }
}
