package org.bdware.sc.units;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.ContractManager;
import org.bdware.sc.conn.Node;
import org.bdware.sc.sequencing.SequencingAlgorithm;

import java.io.Serializable;
import java.util.*;

public class ContractUnitController {
    private static final Logger LOGGER = LogManager.getLogger(ContractUnitController.class);
    private final TrustfulExecutorConnection connection;
    private final Map<String, ContractUnit> units;   //这个节点上<合约id，合约集群>，其中每个ContractUnit中存储了对应合约id的其他节点udp信息
    ContractManager manager;
    SequencingAlgorithmFactory algorithmFactory;

    public ContractUnitController(
            TrustfulExecutorConnection connection,
            ContractManager manager,
            SequencingAlgorithmFactory factory) {
        this.connection = connection;
        this.manager = manager;
        algorithmFactory = factory;
        units = new HashMap<>();
    }

    public void onContractUnitMessage(final ContractUnitMessage cumsg, Node node) {
        ContractUnit unit;
        switch (cumsg.type) {
            case Start:
                ContractUnitStartRequest req = ContractUnitStartRequest.parse(cumsg.content);
                unit = new ContractUnit();
                unit.connection = connection;
                String result = manager.startContract(req.contract);
                unit.contractID = req.contract.getID();
                unit.sequencingAlgorithm = algorithmFactory.create(req, unit);
                unit.node2member = new HashMap<>();
                units.put(req.contract.getID(), unit);

                System.out.println(
                        "[ContractUnitController] startContract:"
                                + result
                                + " isMaster:"
                                + req.isMaster
                                + " cid:"
                                + req.contract.getID());
                break;
            case AddMember:
                LOGGER.debug("contractID:" + cumsg.getContractID());
                unit = units.get(cumsg.getContractID());
                unit.addMember(node, cumsg);   //启动通过在UDPTrustExecutor中遍历memebers，使得这个合约集群中每个节点都有其他节点的UDP信息，便于之后护发UDP消息
                break;
            case Sequencing:
                unit = units.get(cumsg.getContractID());
                ContractUnitMember member = unit.node2member.get(node);
                unit.sequencingAlgorithm.onMessage(member, cumsg.content);
                break;
            case Unknown:
            default:
                //recover start
                ContractUnitStartRequest req2 = ContractUnitStartRequest.parse(cumsg.content);
                unit = new ContractUnit();
                unit.connection = connection;
                unit.contractID = req2.contract.getID();
                unit.sequencingAlgorithm = algorithmFactory.create(req2, unit);
                unit.node2member = new HashMap<>();
                units.put(req2.contract.getID(), unit);

                System.out.println(
                        "[ContractUnitController] startContract:"
                                + " isMaster:"
                                + req2.isMaster
                                + " cid:"
                                + req2.contract.getID());
                break;
        }
    }

    public ContractUnit getContractUnit(String id) {
        return units.get(id);
    }

    public static class ContractUnit implements TrustfulExecutorConnection, Serializable {
        public String contractID;
        public Map<Node, ContractUnitMember> node2member;  //存储和自己在一个合约集群的其他节点信息
        private transient SequencingAlgorithm sequencingAlgorithm;
        private transient TrustfulExecutorConnection connection;

        public synchronized void broadcast(ContractUnitMessage cuMessage) {
            for (Node node : node2member.keySet()) {
                connection.sendMessage(node, cuMessage.toByteArray());
            }
        }

        public String addMember(Node node, ContractUnitMessage msg) {
            ContractUnitMember member;
            try {
                // TODO
                member = new ContractUnitMember(node);
                sequencingAlgorithm.onMessage(member, msg.content);
                node2member.put(node, member);
                return "success";
            } catch (Exception e) {
                e.printStackTrace();
                return "failed";
            }
        }

        @Override
        public void sendMessage(Node node, byte[] msg) {
            ContractUnitMember member = (ContractUnitMember) node;
            ContractUnitMessage unitMsg = new ContractUnitMessage();
            unitMsg.setContractID(contractID);
            unitMsg.type = ContractUnitType.Sequencing;
            unitMsg.content = msg;
            System.out.println("[ContractUnitController] sendMsg:" + unitMsg.getDisplayStr());

            // TODO sig here.
            connection.sendMessage(member.getNode(), unitMsg.toByteArray());
        }
    }

    static class PCInfo {
        public ContractUnitMessage request;
        Set<Integer> prepare;
        Set<Integer> commit;
        boolean isPrePrepareReceived;
        boolean isSendCommit;
        boolean isSendReply;
        List<ContractUnitMessage> buff;

        PCInfo() {
            prepare = new HashSet<>();
            commit = new HashSet<>();
            buff = new ArrayList<>();
            isSendCommit = false;
            isSendReply = false;
            isPrePrepareReceived = false;
        }

        public String getDisplayStr() {
            return "pSize:"
                    + (prepare == null ? "null" : prepare.size())
                    + " cSize:"
                    + (commit == null ? "null" : commit.size())
                    + " isSendCommit:"
                    + isSendCommit
                    + " isSendReply:"
                    + isSendReply
                    + " buffSize:"
                    + buff.size();
        }
    }
}
